********************************************
********   INITIALIZATION PROGRAM   ********
********   FOR DRIVES ON SASI BUS   ********
********************************************
VERS	EQU	'2 '		; December 3, 1982 3:31 klf "INITSASI.ASM"
********************************************
******** MACRO ASSEMBLER DIRECTIVES ********
	MACLIB	Z80
	$-MACRO

********************************************
**  PORTS AND CONSTANTS
********************************************

GPIO	EQU	0F2H		; SWITCH 501

ACK	EQU	00000001B	; SASI STATUS PORT BITS
INT	EQU	00000010B
PER	EQU	00000100B
BUSY	EQU	00001000B
CMND	EQU	00010000B
MSG	EQU	00100000B
POUT	EQU	01000000B
REQ	EQU	10000000B

RUN	EQU	00000000B	; SASI CONTROL COMMANDS
SWRS	EQU	00010000B
INTE	EQU	00100000B
SEL	EQU	01000000B

RECAL	EQU	1		; CONTROLLER COMMANDS
FMT	EQU	4
CHKTRK	EQU	5
RDBL	EQU	8
WRBL	EQU	10
TSTDRV	EQU	0E3H

CPM	EQU	0		; ADDRESSES
BDOS	EQU	5
TIME	EQU	0BH
FCB	EQU	5CH
DMA	EQU	80H
TPA	EQU	100H

DPBL	EQU	21		; MISCELLANEOUS CONSTANTS
MSGLEN	EQU	56

DCTYPE	EQU	3		; LOCATIONS IN MAGIC SECTOR
CBYTE	EQU	4
DDATA	EQU	5
CSTRNG	EQU	13
NPART	EQU	19
DDPB	EQU	47
SECTBL	EQU	20

CONIN	EQU	1		; BDOS CALLS
CONOUT	EQU	2
MSGOUT	EQU	9
OPEN	EQU	15
READF	EQU	20
SETDMA	EQU	26

CR	EQU	0DH		; ASCII CHARACTERS
LF	EQU	0AH
BS	EQU	8
BEL	EQU	7
ESC	EQU	1BH
CTLC	EQU	3
********************************************
	ORG	TPA

START:	LXI	SP,STACK	; LOCAL STACK
	LXI	D,SIGNON
	CALL	PMSG		; TYPE SIGN ON MESSAGE
	CALL	RDFILE		; READ IN HEX FILE
	CALL	GETSEC		; GET MAGIC SECTOR FOR L.U.N. 0
	JC	BDHFIL		; ERROR IF SOMETHING'S WRONG WITH THIS SECTOR
	LXI	D,MENU		; DISPLAY TEXT MESSAGES AND MENU
	CALL	PMSG
	JR	CHOOSE
PROMPT: LXI	D,PMTMSG
PROMT1: CALL	PMSG
	CALL	GETCHR		; GET A CHARACTER FROM USER
	LXI	D,BELMSG
	CPI	CR
	JRNZ	PROMT1
CHOOSE: LXI	D,CHCMSG
BADCHR: CALL	PMSG
	CALL	GETCHR		; INPUT A CHARACTER FROM USER
	LXI	D,CRLF		; EXTRA CR-LF NEEDED BEFORE WARM BOOT
	CPI	CTLC
	JZ	EXIT		; REBOOT ON CONTROL-C
	CPI	'4'		;  OR THE NUMBER 4
	JZ	EXIT 
	LXI	D,BELMSG	; RING BELL IF CHARACTER OUT OF RANGE
	CPI	'R'
	JRZ	SECRET		; "R" IS THE SECRET OPTION
	CPI	'1'
	JC	BADCHR
	CPI	'5'		; OTHERWISE MUST BE 1 THRU 4
	JNC	BADCHR
SECRET: PUSH	PSW		; SAVE THE CHARACTER
	LXI	D,CURPOS	; POSITION CURSOR FOR ERROR MESSAGE, WHILE
	CALL	PMSG		;  SAVING POSITION FOR LOGICAL UNIT MESSAGE
	CALL	GET$PORT	; DETERMINE CONTROLLER BASE PORT ADDRESS
	CALL	RESET		; CLEAN UP
	POP	PSW		; GET CHARACTER BACK
	CPI	'1'		; JUMP TO ROUTINE DEPENDING ON CHARACTER
	JZ	TESTCON
	CPI	'2'
	JZ	DO$IT$ALL
	CPI	'3'
	JZ	TESTDRV
	JMP	DO$CONFIG	; SECRET OPTION - CONFIGURE WITHOUT FORMATTING

TESTCON:
	CALL	CHKXBC		; CHECK FOR XEBEC - NO TEST IF NOT XEBEC
	JNZ	GOTMSG
	CALL	CONCHK		; PERFORM CONTROLLER DIAGNOSTIC
	CALL	RAMCHK		; PERFORM RAM BUFFER DIAGNOSTIC
	JMP	PROMPT

DO$IT$ALL:
	CALL	ASKLUN		; SEE IF USER WANTS A DIFFERENT DRIVE
	CALL	CHKXBC		; SKIP DIAGNOSTICS IF NOT XEBEC
	JNZ	NODIAG
	CALL	CONCHK		; PERFORM CONTROLLER DIAGNOSTIC
	CALL	RAMCHK		; PERFORM RAM BUFFER DIAGNOSTIC
NODIAG: CALL	INSPEC		; INSPECT SECTOR 0 OF DRIVE
	CALL	NER1		; RESET AND INITIALIZE DRIVE
	CALL	FORMAT
	CALL	CHKXBC		; SKIP DIAGNOSTIC IF NOT XEBEC
	JNZ	NODIA1
	CALL	DRVCHK		; PERFORM DRIVE DIAGNOSTIC
NODIA1: CALL	CONFIG		; WRITE MAGIC SECTOR
	CALL	CLRDIR		; CLEAR ALL DIRECTORIES
	JMP	PROMPT

DO$CONFIG:
	CALL	ASKLUN		; SEE IF USER WANTS A DIFFERENT DRIVE
	CALL	CHKXBC		; SKIP DIAGNOSTICS IF NOT XEBEC
	JNZ	NODIA2
	CALL	CONCHK		; PERFORM CONTROLLER DIAGNOSTIC
	CALL	RAMCHK		; PERFORM RAM DIAGNOSTIC
	CALL	DRVCHK		; PERFORM DRIVE DIAGNOSTIC
NODIA2: CALL	NER1		; RESET AND INITIALIZE DRIVE
	CALL	CONFIG
	JMP	PROMPT
 
TESTDRV:
	CALL	CHKXBC		; CHECK FOR XEBEC - NO TEST IF NOT XEBEC
	JNZ	GOTMSG
	CALL	ASKLUN		; SEE IF USER WANTS A DIFFERENT DRIVE
	CALL	DRVCHK		; PERFORM DRIVE DIAGNOSTIC
	JMP	PROMPT

*******************************************************************************
;
;	ROUTINE TO READ 1024 BYTES FROM THE SPECIFIED HEX FILE INTO 'BUFFER'
;

RDFILE:

;
; OPEN HEX FILE

	LDA	FCB+1		; CHECK FOR FILE NAME TYPED IN
	CPI	' '
	LXI	D,NONAME
	JZ	EXIT		; ERROR IF NO NAME GIVEN
	LXI	H,FCB
	LXI	D,HEXFILE
	LXI	B,9
	LDIR			; MOVE TO LOCAL FCB
	LXI	D,HEXFILE
	MVI	C,OPEN
	CALL	BDOS		; OPEN THE HEX FILE
	CPI	255
	JZ	NOHFILE 	; ERROR IF FILE DOES NOT EXIST
	CALL	USFILE

;
; SEARCH FOR BEGINNING OF DRIVER, GET TEXT STRING AND CONTROLLER NUMBER
;
	LXI	H,DMA		; INITIALIZE BYTE COUNT AND ADDRESS
	SHLD	BADR
	XRA	A
	STA	COUNT
NXTLIN: CALL	OPEN$LINE	; FIND BEGINNING OF NEXT LINE
	JC	BDHFIL		; END OF FILE ERROR
	CPI	0C3H		; FIRST BYTE OF DRIVER IS C3
	JRNZ	NXTLIN
	MOV	A,H
	ORA	L		; LOOK FOR ADDRESS OF 0000
	JRNZ	NXTLIN
	MVI	B,9		; SKIP OVER 9 BYTES (THE THREE JMP's AT
SKIP1:	PUSH	B		;  START OF DRIVER)
	CALL	GET$BYTE
	POP	B
	JC	BDHFIL		; END OF FILE ERROR
	DJNZ	SKIP1
	LXI	H,TXTMSG
NXTCHR: CPI	'$'		; '$' MARKS END OF TEXT STRING
	JRZ	DONTXT
	MOV	M,A		; PUT CHARACTER INTO MESSAGE
	INX	H
	PUSH	H
	CALL	GET$BYTE	; GET ANOTHER BYTE
	POP	H
	JC	BDHFIL		; END OF FILE ERROR
	JR	NXTCHR
DONTXT: MVI	B,146		; SKIP OVER 144 BYTES (DISK PARAMETER HEADERS)
SKIP2:	PUSH	B		; 145th BYTE IS CONTROLLER NUMBER
	CALL	GET$BYTE
	POP	B
	JC	BDHFIL		; END OF FILE ERROR
	DJNZ	SKIP2
	CPI	8
	JNC	BDHFIL		; ERROR -- CONTROLLER NUMBER OUT OF RANGE
	STA	CNUM		; SAVE CONTROLLER NUMBER
	ORI	'0'		; MAKE INTO CHARACTER
	STA	CNMMSG		; PUT INTO MESSAGE

; SEARCH FOR A RECORD OF 0000000000. THIS INDICATES END OF DRIVER
;  MODULE. NEXT RECORD IS 'MAGIC SECTOR'
;
NXTBYT: CALL	DREAD		; GET A BYTE FROM THE HEX FILE
	JC	BDHFIL		; END OF FILE ERROR
	CPI	':'		; SEARCH FOR START OF LINE CHARACTER
	JRNZ	NXTBYT
	CALL	DHEX		; FOUND ':'. NOW LOOK FOR '00'
	ORA	A
	JRNZ	NXTBYT

;
; FOUND END OF DRIVER. NOW LOOK FOR TEXT STRING AT BEGINNING OF 'MAGIC SECTOR'
;
FSEMI0: CALL	DREAD		; SEARCH FOR A SEMI-COLON
	JC	BDHFIL		; NO STRING MEANS BAD FILE
	CPI	';'
	JRNZ	FSEMI0
	CALL	DREAD
	CPI	'"'		; NEXT CHARACTER MUST BE A DOUBLE QUOTE
	JNZ	BDHFIL		;  BAD FILE IF NOT
	LXI	H,CNMMSG+3
NXTCH0: CALL	DREAD		; GET CHARACTER FROM HEX FILE
	JC	BDHFIL
	CPI	'"'		; END OF STRING
	JRZ	EOSTR0
	MOV	M,A		; PUT CHARACTER IN MESSAGE
	INX	H
	JR	NXTCH0
EOSTR0: LXI	H,LUNMSG+16	; NOW GET MESSAGES FOR EACH L.U.N.
NXTSTR: PUSH	H
FSEMI:	CALL	DREAD		; SEARCH FOR A SEMI-COLON
	JRC	NOSTR		; END OF FILE => NO MORE STRINGS
	CPI	';'
	JNZ	FSEMI
	CALL	DREAD		; NEXT BYTE IS L.U.N IN ASCII
	JC	BDHFIL		; END OF FILE ERROR
CHKLUN: CMP	M
	JRZ	PUTSTR
	JC	BDHFIL
	CPI	'4'
	JNC	BDHFIL		; ERROR IF L.U.N IS OUT OF RANGE
	POP	H
	LXI	D,MSGLEN
	DAD	D
	PUSH	H
	JR	CHKLUN
PUTSTR: MOV	M,A
	CALL	DREAD		; NEXT CHARACTER MUST BE A DOUBLE QUOTE
	JC	BDHFIL		; END OF FILE ERROR
	CPI	'"'
	JNZ	BDHFIL
	INX	H
	INX	H
NXTCH1: CALL	DREAD
	JC	BDHFIL		; END OF FILE ERROR
	CPI	'"'		; A SECOND DOUBLE QUOTE MARKS END OF STRING
	JRZ	EOSTR
	INX	H
	MOV	M,A		; PUT CHARACTER INTO MESSAGE
	JR	NXTCH1
EOSTR:	POP	H
	LXI	D,MSGLEN	; FIND START OF NEXT STRING
	DAD	D
	JR	NXTSTR
NOSTR:	POP	H
	RET
;
; GETSEC: READ IN THE MAGIC SECTOR FOR THE REQUESTED LOGICAL UNIT
;
GETSEC: XRA	A
	STA	CREC		; START FILE OVER AT RECORD 0
NXTBYT1:CALL	DREAD		; GET A BYTE FROM THE HEX FILE
	CPI	':'		; SEARCH FOR START OF LINE CHARACTER
	JRNZ	NXTBYT1
	CALL	DHEX		; FOUND ':'. NOW LOOK FOR '00'
	ORA	A
	JRNZ	NXTBYT1
FSEMI1: CALL	DREAD		; FOUND END OF DRIVER. LOOK FOR A MAGIC SECTOR
	RC			; EOF => NO MAGIC SECTOR FOR THIS L.U.N.
	CPI	';'		; SEARCH FOR A SEMI-COLON
	JRNZ	FSEMI1
	CALL	DREAD		; GET NEXT CHARACTER
	RC			; EOF => NO MAGIC SECTOR FOR THIS L.U.N.
	MOV	C,A
	LDA	LUN		; SAME AS REQUESTED L.U.N. ?
	CMP	C
	JRNZ	FSEMI1		; LOOK FOR ANOTHER MAGIC SECTOR
; FOUND CORRECT MAGIC SECTOR. NOW READ IT IN.
	CALL	OPEN$LINE	; FIND START OF NEXT LINE AND GET FIRST BYTE
	RC			; END OF FILE ERROR
	LXI	D,2280H
	DSBC	D		; IS ADDRESS 2280H ?
	STC
	RNZ
	CPI	0C3H		; FIRST BYTE SHOULD BE C3
	STC
	RNZ
	STA	BUFFER		; PUT IN BUFFER
	CALL	GET$BYTE
	RC			; END OF FILE ERROR
	CPI	80H		; SECOND BYTE SHOULD BE 80
	STC
	RNZ
	STA	BUFFER+1	; PUT IN BUFFER
	CALL	GET$BYTE
	RC			; END OF FILE ERROR
	CPI	24H		; THIRD BYTE SHOULD BE 24
	STC
	RNZ
	STA	BUFFER+2	; PUT IN BUFFER

;
; NOW READ IN REST OF 'MAGIC SECTOR' AND BOOT LOADER
;
	MVI	C,3		; DO IT THREE TIMES (768 BYTES IN ALL)
	LXI	H,BUFFER+3  
	MVI	B,253
	JR	INLOOP
AGN:	MVI	B,0		; 256 BYTES
INLOOP: PUSH	H
	PUSH	B
	CALL	GET$BYTE	; GET A BYTE
	POP	B
	POP	H
	JC	CKEOF		; END OF FILE OR ERROR
	MOV	M,A		; MOVE IT TO BUFFER
	INX	H
	DJNZ	INLOOP
	MVI	A,3		; AFTER FIRST 256 BYTES (MAGIC SECTOR)
	CMP	C		;  WE NEED TO SKIP 256 BYTES
	JRNZ	NOSKIP
	LXI	H,BUFFER+512
NOSKIP: DCR	C		; GET ANOTHER 256 BYTES ?
	JNZ	AGN
				; IF WE HAVEN'T HIT END OF FILE YET,
				;  SOMETHING'S WRONG (FILE IS TOO LONG)
CKEOF:	LDA	EOF		; IS IT A LEGITIMATE END OF FILE ?
	DCR	A
	STC
	RNZ			; ERROR IF NOT
	LHLD	ADDRS
	LXI	D,2580H 	; BOOT LOADER (STARTS AT 2480H) IS AT LEAST
	ORA	A		;  256 BYTES LONG. ERROR IF END OF FILE
	DSBC	D		;  OCCURS TO SOON.
	RET
******************************************************************************


ASKLUN: LXI	D,QLUMSG	; ASK USER WHAT L.U.N. TO ACCESS
ASKAGN: CALL	PMSG
	CALL	GETCHR
	LXI	D,BELMSG
	CPI	CR		; <CR> MEANS NO CHANGE
	JRNZ	NOTCR		; IF LUN IS A BLANK, THEN <CR> IS INVALID
	LDA	LUN
NOTCR:	CPI	'0'		; CHECK FOR BAD RESPONSE
	JRC	ASKAGN
	CPI	'4'
	JRNC	ASKAGN
	PUSH	PSW		; SAVE THE CHARACTER
	LXI	D,CRLFLF
	CALL	PMSG		; CR-LF-LF TO GET CURSOR TO WHERE IT SHOULD BE
	POP	PSW		; GET CHARACTER BACK
	LXI	H,LUN
	CMP	M
	RZ			; DONE IF SAME L.U.N.
	MOV	M,A		; SET NEW L.U.N.
	SUI	'0'		; SET BITS IN COMMAND BUFFER FOR NEW L.U.N.
	RRC
	RRC
	RRC
	STA	CMBFR+1
	CALL	GETSEC		; GET MAGIC SECTOR
	LXI	D,BMSMSG	; MESSAGE FOR BAD MAGIC SECTOR
	RNC			; ERROR IF CARRY
	MVI	A,' '		; MAGIC SECTOR IN BUFFER IS BAD, SO NO L.U.N.
	STA	LUN		;  SHOULD BE SELECTED. PUTTING A BLANK HERE
	JMP	GOTMSG		;  WILL FORCE USER TO CHOOSE A NEW L.U.N.
				;  NEXT TIME THIS ROUTINE IS CALLED

RAMCHK: MVI	A,11100000B	; SET UP COMMAND CODE
	STA	CMBFR
	JMP	DOIT0		; EXECUTE COMMAND
 
DRVCHK: LXI	D,DRVMSG	; PRINT 'TESTING DRIVE...' MESSAGE
	CALL	PMSG
	MVI	A,RECAL
	STA	CMBFR
	XRA	A		; DISABLE HALF-STEP (WHY ? BECAUSE IT WORKS.)
	STA	CMBFR+5
	LXI	D,TIMMSG
	CALL	PMSG
	LDA	BUFFER+DCTYPE	; LOOK UP EXPECTED TIME VALUE IN TABLE
	MOV	C,A
	LXI	H,ILVTBL
	LXI	D,5
NXTIL0: MOV	A,M
	CPI	0FFH
	JRZ	DFLT1
	CMP	C
	DAD	D
	JRNZ	NXTIL0
	DCX	H
	MOV	E,M
	MVI	D,0
	SDED	COUNTR	
	LXI	D,EXTMSG	; EXPECTED TIME MESSAGE
	CALL	PMSG
	CALL	PRTRK		; EXPECTED TIME VALUE
	LXI	D,DASH
	CALL	PMSG
	LHLD	COUNTR
	INX	H
	INX	H
	SHLD	COUNTR
	CALL	PRTRK
	LXI	D,RPAREN
	CALL	PMSG
DFLT1:	LXI	H,0
	SHLD	COUNTR		; ZERO COUNTER FOR TIMER
	LXI	D,COLON 
	CALL	PMSG		; "TIMER" MESSAGE
	CALL	WAKE1 
	CALL	OUTCOM
	CALL	CHKTIM		; CHECK STATUS AND PRINT TIMER TICKS
	CALL	NER1		; RESET AND INITIALIZE DRIVE
	MVI	A,TSTDRV	; SET UP COMMAND CODE
	STA	CMBFR
	LDA	BUFFER+CBYTE	; CONTROL BYTE
	ORI	10000000B	; CODE TO DISABLE RETRIES
	STA	CMBFR+5
	CALL	WAKE1
	CALL	OUTCOM		; EXECUTE THE COMMAND
	LXI	D,NRDYMSG
	JNZ	GOTMSG
	CALL	CHKTIM		; CHECK STATUS AND PRINT TIMER TICKS
	JMP	CHKERR

CONCHK: MVI	A,11100100B	; SET UP COMMAND CODE
	STA	CMBFR
	LXI	D,CONMSG	; SET UP MESSAGE

DOIT:	CALL	PMSG		; OUTPUT THE APPROPRIATE MESSAGE
DOIT0:	CALL	WAKE1		; GET CONTROLLER'S ATTENTION
	CALL	OUTCOM		; OUTPUT THE COMMAND
	LXI	D,NRDYMSG
	JNZ	GOTMSG
	CALL	CHK$STAT	; CHECK THE STATUS
CHKERR: LXI	D,OKMSG 	; SET UP MESSAGE FOR NO ERROR
	JZ	NO$ERR		; DONE IF NO ERROR
	MVI	A,00000011B	; SET UP REQUEST SENSE COMMAND
	STA	CMBFR
	CALL	GETCON		; EXECUTE COMMAND
	CALL	OUTCOM
	CALL	IN$SENSE	; INPUT THE 4 SENSE BYTES
	CALL	CHK$STAT	; INPUT STATUS COMPLETION BYTES
	LDA	ERSTAT
	ANI	00110000B	; GET ERROR TYPE
	MVI	B,7
	LXI	H,TYPE0
	CPI	00000000B
	JZ	GOTTYPE 	; IT'S TYPE 0
	MVI	B,11
	LXI	H,TYPE1
	CPI	00010000B
	JZ	GOTTYPE 	; IT'S TYPE 1
	LXI	H,TYPE2
	MVI	B,2
	CPI	00100000B
	JZ	GOTTYPE 	; IT'S TYPE 2
	LXI	H,TYPE3
	MVI	B,3		; IT'S TYPE 3
GOTTYPE:LDA	ERSTAT
	ANI	00001111B	; GET ERROR CODE
	CMP	B
	LXI	D,UNDEFMSG	; UNDEFINED ERROR IF CODE OUT OF RANGE 
	JNC	GOTMSG		;  FOR ERROR TYPE
	RLC			; INDEX INTO MESSAGE ADDRESS TABLE FOR 
	MVI	D,0		;  ERROR TYPE
	MOV	E,A
	DAD	D
	MOV	E,M		; ERROR MESSAGE ADDRESS TO REG. D
	INX	H
	MOV	D,M
	JMP	GOTMSG



IN$SENSE:
	LXI	H,ERSTAT
	MVI	B,4
	LDA	BASE$PORT
	MOV	D,A		; STORE DATA PORT ADDRESS IN REG. D
	INR	A
	MOV	E,A		; CONTROL PORT ADDRESS TO REG. E
INCONT: MOV	C,E		; INPUT FROM CONTROL PORT
	INP	A
	ANI	(REQ  OR MSG OR CMND OR POUT)
	CPI	REQ
	JNZ	INCONT
	MOV	C,D		; INPUT FROM DATA PORT
	INP	A
	MOV	M,A
	INX	H
	DJNZ	INCONT
	RET
 
;
;	ERROR MESSAGE ADDRESS TABLES
;
TYPE0:	DW	OKMSG
	DW	MSG01
	DW	MSG02
	DW	MSG03
	DW	MSG04
	DW	MSG05
	DW	MSG06
	DW	MSG07
	DW	MSG08
	DW	UNDEFMSG
	DW	MSG0A
	DW	MSG0B

TYPE1:	DW	MSG10
	DW	MSG11
	DW	MSG12
	DW	MSG13
	DW	MSG14
	DW	MSG15
	DW	UNDEFMSG
	DW	MSG17
	DW	MSG18
	DW	MSG19
	DW	MSG1A

TYPE2:	DW	MSG20
	DW	MSG21
	DW	MSG22

TYPE3:	DW	MSG30
	DW	MSG31
	DW	MSG32

; ERROR MESSAGES

OKMSG:	DB	ESC,'j  No Errors',ESC,'k$' ; SO THE MESSAGE WON'T APPEAR TWICE
MSG01:	DB	'  NO INDEX SIGNAL$'
MSG02:	DB	'  SEEK NOT COMPLETED$'
MSG03:	DB	'  DRIVE FAULT$'
MSG04:	DB	'  DRIVE NOT READY$'
MSG05:	DB	'  DRIVE NOT SELECTED$'
MSG06:	DB	'  TRACK 0 NOT FOUND (Check Configuration Jumpers)$'
MSG07:	DB	'  MULTIPLE DRIVES SELECTED$'
MSG08:	DB	'  NO ADDRESS ACKNOWLEDGE FROM DP100$'
MSG0A:	DB	'  ADDR ACK/INV FAULT$'
MSG0B:	DB	'  NO DATA TRANSITION$'
MSG10:	DB	'  ID READ ERROR$'
MSG11:	DB	'  UNCORRECTABLE DATA ERROR$'
MSG12:	DB	'  NO ID ADDRESS MARK$'
MSG13:	DB	'  NO DATA ADDRESS MARK$'
MSG14:	DB	'  SECTOR NOT FOUND$'
MSG15:	DB	'  SEEK ERROR$'
MSG17:	DB	'  WRITE PROTECTED SECTOR$'
MSG18:	DB	'  CORRECTABLE DATA ERROR$'
MSG19:	DB	'  BAD TRACK$'
MSG1A:	DB	'  FORMAT ERROR$'
MSG20:	DB	'  INVALID COMMAND$'
MSG21:	DB	'  ILLEGAL DISK ADDRESS$'
MSG22:	DB	'  ILLEGAL FUNCTION FOR CURRENT DRIVE TYPE$'
MSG30:	DB	'  RAM ERROR$'
MSG31:	DB	'  PROGRAM MEMORY CHECKSUM ERROR$'
MSG32:	DB	'  ECC POLYNOMIAL ERROR$'
UNDEFMSG: DB	'  UNDEFINED ERROR$'

PAGE



;	DATA BUFFERS AND STORAGE
;
ERSTAT: DB	0,0,0,0
******************************************************************************
;
; INSPECT SECTOR 0, FORMAT, AND WRITE 'MAGIC SECTOR'
;

;
; TRY TO READ SECTOR 0 OF THE DISK TO SEE IF IT HAS BEEN
;  PREVIOUSLY CONFIGURED.
;
INSPEC: LXI	D,INSMSG	; 'INSPECTING DRIVE' MESSAGE
	CALL	PMSG
	XRA	A
	STA	CMBFR		; FIRST TEST DRIVE READY
	CALL	WAKE1
	CZ	OUTCOM
	CZ	CHK$STAT
	LXI	D,DNRMSG	; DRIVE NOT READY ERROR MESSAGE
	JNZ	GOTMSG
	MVI	A,RDBL		; SET UP READ COMMAND
	STA	CMBFR
	XRA	A
	STA	CMBFR+1 	; SECTOR 0 ADDRESS
	STA	CMBFR+2
	STA	CMBFR+3
	INR	A
	STA	CMBFR+4 	; READ 1 SECTOR
	LDA	BUFFER+CBYTE
	STA	CMBFR+5
	CALL	WAKE1
	CZ	OUTCOM
	CZ	READ
	CZ	CHK$STAT
	LXI	D,NODMSG	; MESSAGE FOR DRIVE NOT PREVIOUSLY CONFIGURED
	JNZ	NO$ERR		; IF READ ERROR, GO ON AND FORMAT DRIVE
	LXI	H,RDBUF
	MOV	A,M
	CPI	0C3H		; IF DRIVE HAS BEEN PREVIOUSLY CONFIGURED,
	JNZ	NO$ERR		;  FIRST BYTE SHOULD BE C3
	INX	H
	MOV	A,M
	CPI	80H		; SECOND BYTE SHOULD BE 80
	JNZ	NO$ERR
	INX	H
	MOV	A,M
	CPI	24H		; THIRD BYTE SHOULD BE 24
	JNZ	NO$ERR
	LXI	D,PREMSG
	CALL	PMSG
	CALL	GETCHR		; GET CHARACTER FROM USER
	CPI	'Y'
	LXI	D,NFMMSG	; IF 'N' PRINT MESSAGE AND GO BACK TO PROMPT
	JNZ	GOTMSG
	LXI	D,DDSMSG	; IF 'Y' PRINT MESSAGE AND GO ON
	JMP	NO$ERR

;
; FORMAT THE DISK
;
FORMAT: LXI	D,FMTDSK
	CALL	PMSG
	LXI	H,BUFFER+DDATA	; CALCULATE MAXIMUM TRACK NUMBER
	MOV	D,M		;  NUM. CYLINDERS x NUM. HEADS
	INX	H
	MOV	E,M
	INX	H
	MOV	B,M
	LXI	H,0
MULT:	DAD	D
	DJNZ	MULT
	DCX	H
	SHLD	MAXTRK		; SAVE FOR USE BY VERIFY ROUTINE
	MVI	A,FMT		; FORMAT DRIVE COMMAND
	STA	CMBFR
	LXI	D,TIMMSG
	CALL	PMSG
	LDA	BUFFER+DCTYPE	; GET DRIVE/CONTROLLER TYPE
	MOV	C,A
	LXI	H,ILVTBL	; LOOK UP INTERLEAVE FACTOR
	LXI	D,5
NXTILV: MOV	A,M
	CPI	0FFH		; FF MARKS END OF TABLE
	JRZ	DFLT0		; IF NOT IN TABLE, DEFAULT TO 0
	CMP	C
	DAD	D
	JRNZ	NXTILV		; CHECK NEXT TABLE ENTRY
	DCX	H
	DCX	H
	MOV	A,M
	STA	SPT		; SAVE SECTORS PER TRACK FOR FUTURE USE
	DCX	H
	MOV	E,M
	MVI	D,0
	SDED	COUNTR		; EXPECTED TIME VALUE TO COUNTER
	DCX	H
	MOV	A,M
	STA	CMBFR+4 	; PUT INTERLEAVE FACTOR INTO COMMAND BUFFER
	LXI	D,EXTMSG
	CALL	PMSG		; PRINT EXPECTED TIME MESSAGE
	CALL	PRTRK		; PRINT EXPECTED TIME VALUE (LOWER BOUND)
	LHLD	COUNTR
	INX	H
	INX	H
	SHLD	COUNTR		; UPPER BOUND = LOWER BOUND + 2
	LXI	D,DASH		; PRINT A DASH
	CALL	PMSG
	CALL	PRTRK		; PRINT EXPECTED TIME (UPPER BOUND)
	LXI	D,RPAREN
	CALL	PMSG
DFLT0:	LXI	H,0
	SHLD	COUNTR		; ZERO COUNTER FOR FORMAT TIMER
	LXI	D,COLON
	CALL	PMSG		; PRINT FORMAT TIMER MESSAGE
	CZ	GETCON		; GET CONTROLLER'S ATTENTION
	CZ	OUTCOM		; OUTPUT THE COMMAND
	CZ	CHKTIM		; CHECK THE STATUS AND PRINT TIMER TICKS
	CALL	CHKMRX		; MEMOREX CONTROLLER DOESN'T SUPPORT 
	JZ	NER1		;  CHECK TRACK FORMAT COMMAND
	MVI	A,CHKTRK	; CHECK TRACK FORMAT COMMAND
	STA	CMBFR
	LXI	H,0
	SHLD	COUNTR		; ZERO COUNTER FOR TRACKS
	SHLD	ERRCNT
	LXI	D,VERTRK	; OUTPUT 'VERIFYING TRACK NUMBER' MESSAGE
	CALL	PMSG
NXTTRK: CALL	PRTRK		; PRINT THE TRACK NUMBER
	CALL	GETCON		; GET CONTROLLER'S ATTENTION
	CZ	OUTCOM		; OUTPUT THE COMMAND
	CZ	CHK$STAT	; CHECK THE STATUS
	CNZ	TRKERR
	CALL	INCRE		; INCREMENT TO NEXT TRACK
	JRNC	NXTTRK
DONE:	LHLD	ERRCNT
	MOV	A,H
	ORA	L
	JZ	NER1		; NO MESSAGES IF NO ERRORS
	SHLD	COUNTR		; NUMBER OF BAD TRACKS TO COUNTR
	LXI	D,NBTMSG
	CALL	PMSG		; MESSAGE
	CALL	PRTRK		; PRINT NUMBER
	JMP	GOT1		;  AND EXIT

TRKERR: LHLD	ERRCNT		; SEE IF FIRST VERIFY ERROR
	MOV	A,H
	ORA	L
	JRNZ	NOT$1ST
	LXI	D,FBTMSG	; 'FIRST BAD TRACK' MESSAGE
	CALL	PMSG
	LXI	D,LBTMSG	; 'LAST BAD TRACK' MESSAGE
	CALL	PMSG
	LXI	H,0
NOT$1ST:INX	H
	SHLD	ERRCNT		; INCREMENT THE ERROR COUNT
	LXI	D,CURLBT	; POSITION CURSOR TO LAST BAD ... MESSAGE
	CALL	PMSG
	CALL	PRTRK
	LXI	D,CURVER	; REPOSITION CURSOR TO VERIFYING ... MESSAGE
	CALL	PMSG
	RET

;
; DISK SUCCESSFULLY FORMATTED. NOW WRITE MAGIC SECTOR
;
CONFIG: LXI	D,WRMSG
	CALL	PMSG
	XRA	A		; WRITE TO SECTOR 0
	STA	CMBFR+1
	STA	CMBFR+2
	STA	CMBFR+3
	MVI	A,8		; NUMBER OF LOGICAL SECTORS TO WRITE
	LXI	H,BUFFER	; WRITE BUFFER -- MAGIC SECTOR IS HERE
	CALL	WRBLK		; WRITE THE BLOCK
	JNZ	GOTMSG		; REPORT ERROR
	JMP	NER1

;
; CLEAR ALL DIRECTORIES
;
CLRDIR: LXI	D,CLRMSG	; "CLEARING DIRECTORIES ..." MESSAGE
	CALL	PMSG
	MVI	A,WRBL 
	STA	CMBFR		; CODE TO WRITE TO DISK
	LDA	BUFFER+NPART	; NUMBER OF PARTITIONS
	MOV	B,A
CLRONE: PUSH	B
	LXIX	BUFFER+DDPB-DPBL; DPB TABLE
	LXIY	BUFFER+SECTBL-3 ; PARTITION TABLE
	LXI	D,DPBL
NXTDPB: DADX	D		; POINT IX TO A DPB
	INXIY
	INXIY
	INXIY
	DJNZ	NXTDPB
	LDX	B,+13		; (B) = OFFSET
	LDX	E,+0
	LDX	D,+1		; (DE) = SPT
	LXI	H,0
MULT2:	DAD	D		; SYSSIZ=OFFSET*SPT
	DJNZ	MULT2
	LDY	C,+0
	LDY	D,+1		; (DE) = PARTITION OFFSET
	LDY	E,+2
	DAD	D		; DIRADR = SYSSIZ + PARTITION OFFSET
	JNC	NOCARR
	INR	C
NOCARR: XCHG
	CALL	SECTRN		; TRANSLATE TO PHYSICAL SECTORS
	MOV	A,C
	STA	CMBFR+1
	MOV	A,D
	STA	CMBFR+2
	MOV	A,E
	STA	CMBFR+3 	; PUT SECTOR NUMBER INTO COMMAND BUFFER
	MVI	A,1		; WRITE 1 SECTOR
	STA	CMBFR+4
	LDX	H,+9
	LDX	L,+10		; (HL) = ALV
	XRA	A
ROTHL:	ANA	A		; CLEAR CARRY
	RALR	L		; SEE HOW MANY BITS ARE SET
	RALR	H		; THIS IS THE NUMBER OF BLOCKS IN THE DIRECTORY
	JNC	GOTBLK		;  AND THE BLOCK NUMBER FOR THE CCP FILE
	INR	A
	JR	ROTHL
GOTBLK: MOV	B,A		; (B)=NUMBLKS
	LDX	C,+3		; (A) = BSM
	INR	C		; (C) = SPB
	MOV	A,C
	XRA	A
MULT1:	ADD	C		; DIRSIZ=NUMBLKS*SPB
	DJNZ	MULT1
	MOV	E,A
	MVI	D,0
	MVI	C,0
	CALL	SECTRN		; TRANSLATE TO PHYSICAL SECTORS
	MOV	A,E
CLRMOR: PUSH	D		; (E) = NUMBER OF PHYSICAL SECTORS TO WRITE
	CALL	GETCON
	CZ	OUTCOM
	LXI	H,EMTDIR	; BUFFER FULL OF E5's
	CZ	WRITE		; WRITE THE SECTOR
	CZ	CHK$STAT
	LXI	D,WERMSG
	JNZ	GOTMSG
	LXI	H,CMBFR+3	; INCREMENT ADDRESS TO NEXT SECTOR
	INR	M
	JNZ	NOCARY
	DCX	H
	INR	M
	JNZ	NOCARY
	DCX	H
	INR	M
NOCARY: POP	D
	DCR	E		; MORE TO WRITE ?
	JNZ	CLRMOR
	POP	B
	DCR	B
	JNZ	CLRONE
	JMP	NER1

*******************************************************************************
; SUBROUTINES
*******************************************************************************

;
; SUBROUTINE TO GET A CHARACTER FROM THE CONSOLE
;
GETCHR: MVI	C,CONIN
	CALL	BDOS
	CPI	'a' ! RC	; RETURN IF BELOW LOWER CASE A
	CPI	'z'+1
	RNC			; RETURN IF ABOVE LOWER CASE Z
	ANI 01011111B		; LOWER CASE LETTER, MAKE UPPER
	RET

CHKXBC: LDA	BUFFER+DCTYPE	; GET DRIVE/CONTROLLER TYPE BYTE
	ANI	11100000B	; ISOLATE CONTROLLER TYPE BITS
	LXI	D,NXBMSG	; XEBEC CODE IS 000
	RET

CHKMRX: LDA	BUFFER+DCTYPE	; GET DRIVE/CONTROLLER TYPE BYTE
	ANI	11100000B	; ISOLATE CONTROLLER TYPE BITS
	CPI	01000000B	; MEMOREX CODE IS 010
	RET

;
; GOTMSG: PRINT ERROR MESSAGE, AND RETURN TO PROMPT
; GOT1:   RETURN TO PROMPT
; NO$ERR: PRINT ERROR MESSAGE, CLEAN UP
; NER1:   CLEAN UP
;
GOTMSG: CALL	PMSG  
GOT1:	LXI	SP,STACK	; RE-SET STACK
	JMP	PROMPT		; AND RETURN TO PROMPT

NO$ERR: CALL	PMSG		; PRINT ERROR MESSAGE
NER1:	CALL	RESET
	CALL	INIT$DRIVE
	RET			; RETURN TO PROMPT

;
; OPEN$LINE: READS BEGINNING OF LINE FROM HEX FILE
; GET$BYTE: GETS ONE BYTE FROM HEX FILE
;
OPEN$LINE:
	CALL	DREAD		; GET A BYTE FROM FILE
	RC			; END OF FILE ERROR
	CPI	':'		; SEARCH FOR START OF LINE CHARACTER
	JNZ	OPEN$LINE
	XRA	A
	STA	CHKSUM		; INITIALIZE CHECKSUM
	CALL	DHEX		; GET FIRST BYTE (NUMBER OF DATA BYTES IN LINE)
	STA	BYTEC
	ORA	A
	STC
	JRNZ	NEOF		; A LINE OF 0 INDICATES END OF FILE
	MVI	A,1
	STA	EOF		; FLAG LEGITIMATE END OF FILE
	RET
NEOF:	CALL	DHEX		; GET HI BYTE OF ADDRESS
	PUSH	PSW
	CALL	DHEX		; GET LO BYTE OF ADDRESS
	MOV	L,A
	POP	PSW	
	MOV	H,A
	SHLD	ADDRS		; SAVE STARTING ADDRESS OF LINE
	CALL	DHEX		; FILL BYTE (SHOULD BE ZERO)
GET$BYTE:
	LDA	BYTEC
	ORA	A
	JZ	OPEN$LINE	; IF LINE IS FINISHED, START A NEW LINE
	CALL	DHEX		; GET DATA BYTE
	PUSH	PSW
	LDA	BYTEC		; COUNT ONE BYTE
	DCR	A
	STA	BYTEC
	JNZ	ONLINE
	CALL	DHEX		; IF AT END OF LINE VERIFY CHESUM BYTE
	LDA	CHKSUM
	ORA	A
	JZ	ONLINE		; CHECKSUM OK
	POP	PSW
	STC			; CHECKSUM ERROR
	RET
ONLINE: POP	PSW		; RETURN DATA BYTE IN ACCUMULATOR
	ORA	A
	LHLD	ADDRS
	PUSH	H		; RETURN ADDRESS OF BYTE IN HL
	INX	H
	SHLD	ADDRS		; SAVE NEXT PREDICTED ADDRESS
	POP	H
	RET

;
; SUBROUTINE TO READ TWO HEX DIGITS FROM FILE AND ASSEMBLE THE BYTE VALUE
;
DHEX:	CALL	DREAD		; READ A BYTE
	CALL	ATOH		; CONVERT FROM ASCII TO HEX
	RLC			; MOVE TO HIGH NYBBLE
	RLC
	RLC
	RLC
	MOV	B,A
	PUSH	B		; SAVE IT
	CALL	DREAD		; READ NEXT BYTE
	CALL	ATOH		; CONVERT FROM ASCII TO HEX
	POP	B		; RETRIEVE HIGH NYBBLE
	ORA	B		; PUT THEM TOGETHER
	MOV	B,A
	LDA	CHKSUM
	ADD	B		; ADD BYTE TO CHEKSUM
	STA	CHKSUM
	MOV	A,B
	RET

;
; SUBROUTINE TO EXTRACT ONE BYTE FROM DMA BUFFER, READING ANOTHER RECORD
;  IF NEEDED
;
DREAD:	PUSH	H
	LHLD	BADR
	LDA	COUNT		; NEED TO READ ANOTHER RECORD ?
	DCR	A
	JP	NOREAD
	LXI	D,HEXFILE
	MVI	C,READF
	CALL	BDOS		; READ FROM HEX FILE
	ORA	A
	STC			; FLAG END OF FILE ERROR
	JRNZ	EOFERR
	MVI	A,127		; REINITIALIZE COUNT
	LXI	H,DMA
NOREAD: STA	COUNT
	MOV	A,M		; RETURN BYTE IN ACCUMULATOR
	INX	H
	SHLD	BADR		; ADDRESS OF NEXT BYTE
	ORA	A		; CLEAR CARRY -- INDICATES NO ERROR
EOFERR: POP	H
	RET

;
; SUBROUTINE TO CONVERT ASCII CHARACTER TO HEX VALUE
;
ATOH:	ADI	0C6H
	JC	ALPH
	ADI	7
ALPH:	ADI	3
	ANI	0FH
	RET

RESET:	LDA	BASE$PORT
	INR	A
	MOV	C,A		; CONTROL PORT TO REG. C
	MVI	A,RUN		; CLEARS ANY PREVIOUS CONTROLLER ACTIVITY
	OUTP	A
	MVI	A,SWRS
	OUTP	A		; RESET CONTROLLER
	RET

INIT$DRIVE:
	CALL	CHKXBC		; INITIALIZE ONLY IF XEBEC CONTROLLER
	JRNZ	NOTXBC
	MVI	A,0CH		; INITIALIZE DRIVE CHARACTERISTICS COMMAND
	STA	CMBFR
	CALL	GETCON		; GET CONTROLLER'S ATTENTION
	CZ	OUTCOM		; OUTPUT COMMAND
	LXI	H,BUFFER+DDATA	; DRIVE CHARACTERISTIC DATA
	MVI	B,8		; 8 BYTES LONG
	MVI	E,(REQ OR POUT OR BUSY)
	CZ	OUTCM1		; OUTPUT THE DATA
	CZ	CHK$STAT	;  AND CHECK STATUS
	RNZ
NOTXBC: LXI	H,BUFFER+CSTRNG ; NOW SEND THE DRIVE ASSIGNMENT COMMAND
	LXI	D,CMBFR 	;  (FOR DATA PERIPHERALS CONTROLLERS)
	LXI	B,6
	LDIR
	CALL	GETCON
	CZ	OUTCOM
	CZ	CHK$STAT
	LDA	BUFFER+CBYTE	; CONTROL BYTE FOR FUTURE COMMANDS
	STA	CMBFR+5
	RET

WRBLK:	PUSH	H		; SAVE BUFFER ADDRESS
	MOV	E,A
	MVI	D,0
	MVI	C,0
	CALL	SECTRN		; TRANSLATE LOGICAL TO PHYSICAL SECTOR
	MOV	A,E
	STA	CMBFR+4 	; NUMBER OF SECTORS TO WRITE
	MVI	A,WRBL		; WRITE COMMAND TO COMMAND BUFFER
	STA	CMBFR
	CALL	GETCON		; SETUP CONTROLLER
	CZ	OUTCOM		; AND OUTPUT THE COMMAND
	POP	H		; WRITE BUFFER TO HL
	CZ	WRITE		; DO WRITE
	CZ	CHK$STAT	; CHECK THE BUS RESPONSE
	LXI	D,WERMSG	; ERROR MESSAGE (IN CASE IT'S NEEDED)
	RET

SECTRN: 
	LDA	BUFFER+DDPB+15	; SECTOR SIZE (FROM 'MAGIC SECTOR')
	ANI	3
	RZ 
ROTATE: ANA	A		; CLEAR CARRY
	RARR	C
	RARR	D
	RARR	E
	DCR	A
	JRNZ	ROTATE
	RET
;
; SUBROUTINE TO DETERMINE BASE PORT ADDRESS. PSW/Z RESET ON ERROR.
;
GET$PORT:
	MVI	B,7CH
	IN	GPIO
	ANI	00000011B
	CPI	00000010B
	JRZ	GOTPRT
	MVI	B,78H
	IN	GPIO
	ANI	00001100B
	CPI	00001000B
	LXI	D,PRTMSG1
	JNZ	GOTMSG
GOTPRT: MOV	A,B
	STA	BASE$PORT
	CALL	RESET
	MVI	A,RECAL
	STA	CMBFR
	CALL	GETCON
	RZ
	LDA	BASE$PORT
	XRI	00000100B	; SWITCH PORT NUMBER
	STA	BASE$PORT
	CALL	RESET
	CALL	GETCON		;  AND TRY AGAIN
	LXI	D,NRDYMSG	; NOT READY IF IT WON'T WORK WITH EITHER PORT
	JNZ	GOTMSG
	CALL	RESET
	LXI	D,PRTMSG2	; MESSAGE FOR WRONG PORT SELECTED
	JMP	GOTMSG

;
; SUBROUTINE TO WAKE UP CONTROLLER, RESETTING IT IF NECESSARY
;
WAKE1:	CALL	GETCON
	RZ
	CALL	RESET
	RNZ
	CALL	GETCON
	RET
;
; SUBROUTINE TO GET THE BUS' ATTENTION
;
GETCON:
	LDA	BASE$PORT
	MOV	C,A
	INR	C		; CONTROL PORT ADDRESS TO REG. C
	MVI	B,0		; TIMER COUNTER
GETCN1: DCR	C		; DATA PORT
	XRA	A
	OUTP	A		; CLEAR DATA REGISTER
	LDA	CNUM		; CONTROLLER NUMBER
	ANI	7
	INR	C
	INR	C		; -SENC- PORT
	OUTP	A
	DCR	C		; CONTROL PORT
	INP	A		; READ CONTROL PORT
	ANI	BUSY
	JRZ	GETCN2
	DJNZ	GETCN1
	DCR	B		; RESET PSW/Z TO INDICATE ERROR
	RET
GETCN2: 
	MVI	A,SEL
	OUTP	A		; WAKE UP CONTROLER
	MVI	B,0
GETCN3:
	INP	A
	ANI	BUSY
	JRNZ	GETCN4
	DJNZ	GETCN3
	DCR	B		; RESET PSW/Z TO INDICATE ERROR
	RET
GETCN4: 
	MVI	A,RUN
	OUTP	A
	XRA	A		; NO ERROR
	RET

;
; SUBROUTINE TO OUTPUT A COMMAND TO SASI BUS
;
OUTCOM:
	LXI	H,CMBFR
	MVI	B,6		; COMMAND IS 6 BYTES LONG
	LDA	BASE$PORT
	MOV	C,A		; AND THE OUTPUT PORT NUMBER
	INR	A
	MOV	D,A		; STORE CONTROL PORT ADDRESS IN REG. D
	MVI	E,(REQ OR CMND OR POUT OR BUSY)
OUTCM1: PUSH	B
	MVI	B,16		; SET LOOP COUNTER
	MOV	C,D		; CONTROL PORT ADDRESS TO REG. C
OUTLOP: INP	A
	ANI	(REQ OR CMND OR POUT OR BUSY)
	CMP	E
	JRZ	OUTOK
	DJNZ	OUTLOP
	DCR	B
	POP	B
	RET
OUTOK:	POP	B		; RETURNS DATA PORT ADDRESS TO REG. C
	OUTI			; OUTPUT COMMAND BYTE
	JNZ	OUTCM1
	XRA	A
	RET

;
; SUBROUTINE TO WRITE DATA TO DISK
;
WRITE:	LDA	BASE$PORT
	MOV	C,A		; DATA PORT ADDRESS TO REG. C
WRMORE: INR	C		; INCREMENT TO CONTROL PORT
CKWRIT: INP	A		; FIRST CHECK FOR DRIVE READY
	ANI	(CMND OR BUSY OR REQ OR POUT)
	CPI	(CMND OR BUSY OR REQ)  ; IF POUT DROPS,


	RZ			       ;  WE ARE INTO STATUS PHASE
	ANI	(CMND OR BUSY OR REQ)
	CPI	(BUSY OR REQ)	; WHEN CMND DROPS, SEEK IS COMPLETE, AND WE ARE
	JRNZ	CKWRIT		;  READY TO WRITE 1 SECTOR
	DCR	C		; DATA PORT ADDRESS TO REG. C
MORE:	MVI	B,128
	OUTIR
	JR	WRMORE		; SEE IF THERE'S ANOTHER SECTOR TO READ IN

READ:	LXI	H,RDBUF
	LDA	BASE$PORT
	MOV	C,A
RDMORE: INR	C
CKREAD: INP	A
	ANI	(CMND OR BUSY OR REQ OR POUT)
	CPI	(CMND OR BUSY OR REQ)
	RZ
	ANI	(CMND OR BUSY OR REQ)
	CPI	(BUSY OR REQ)
	JRNZ	CKREAD
	DCR	C
	MVI	B,128
	INIR
	JR	RDMORE

;
; SUBROUTINE TO CHECK STATUS OF READ OR WRITE
;
CHK$STAT:
	LXI	H,STAT		; STATUS BUFFER
	LDA	BASE$PORT
	MOV	D,A		; DATA PORT ADDRESS STORED IN REG. D
	INR	A
	MOV	E,A		; CONTROL PORT ADDRESS STORED IN REG. E
	JR	CHK01
CHKNXT: MOV	C,D		; INPUT FROM DATA PORT
	INP	A
	MOV	M,A		; SAVE IN MEMORY
CHK01:	MOV	C,E		; INPUT FROM CONTROL PORT
	INP	A
	ANI	(MSG OR REQ OR CMND OR POUT)
	CPI	(REQ OR CMND)
	JRZ	CHKNXT
	CPI	(MSG OR REQ OR CMND)
	JRNZ	CHK01
	MOV	C,D		; INPUT FROM DATA PORT
	INP	A		; GET FINAL BYTE
	MOV	A,M		; AND THROW IT AWAY, GET STATUS
	ANI	03		; EITHER BIT SET IS AN ERROR
	RET

CHKTIM: 			; THIS ROUTINE CHECKS WHAT'S UP
	LXI	H,STAT
	LDA	BASE$PORT
	MOV	D,A		; STORE DATA PORT ADDRESS IN REG. D
	INR	A
	MOV	E,A		; STORE CONTROL PORT ADDRESS IN REG. E
	JR	CHK02
CHKN1:	MOV	C,D		; INPUT FROM DATA PORT
	INP	A
	MOV	M,A		; SAVE DATA IN MEMORY
CHK02:	CALL	PRTIM		; PRINT TIMER TICKS
	MOV	C,E		; INPUT FROM CONTROL PORT
	INP	A
	ANI	(REQ OR MSG OR CMND OR POUT)
	CPI	(REQ OR CMND)
	JRZ	CHKN1
	CPI	(REQ OR MSG OR CMND)
	JRNZ	CHK02		; END OF TRANSMISSION CHECK
	MOV	C,D		; INPUT FROM DATA PORT
	INP	A		; GET FINAL BYTE
	MOV	A,M
	ANI	03		; EITHER BIT SET IS AN ERROR
	RET

;
;	This routine prints a series of dots on the CRT. These dots are
;	printed at timed intervals and are not related to the formatting
;	of the disk. They are only printed to tell the user the program
;	is still running ok.
;
PRTIM:	PUSH	H		; save the registers
	PUSH	D
	PUSH	B
	PUSH	PSW
	LHLD	TIME		; get the time (tic-counter)
	MOV	A,L		; and see if we have had 256 tics
	ORA	A		; yet. L=0 if we have.
	JNZ	NOTICK
	MOV	A,H		; now lets check the H register.
	ANI	3		; we will check the 2 LSB's only
	JNZ	NOTICK		; which will give us approximatly
	LHLD	COUNTR		; 2 seconds between TICKS
	INX	H
	SHLD	COUNTR
	CALL	PRTRK
WMORE:	LHLD	TIME		; here we wait until the next increment
	MOV	A,H		; of the H register (1/2 second), so
	ANI	3		; we are not faster than the counter
	JZ	WMORE
	NOP
NOTICK: POP	PSW		; restore the registers
	POP	B
	POP	D
	POP	H
	RET

;	this routine prints the track number at the cursor position
;	saved at the terminal by 'VERTRK'.
;
PRTRK:	PUSH	H
	PUSH	D
	PUSH	B
	PUSH	PSW
	LXI	D,TRKNUM
	LHLD	COUNTR
	PUSH	D		; convert (HL) to ASCII-decimal (5 digit,
	LXI	B,10000 	; Leading zeros blanked)
	CALL	DECX
	LXI	B,1000
	CALL	DECX
	LXI	B,100
	CALL	DECX		; get 100's digit
	LXI	B,10
	CALL	DECX		; get 10's digit
	MOV	A,L		; get 1's digit (remainder)
	ORI	'0'		; convert to ASCII
	STAX	D		; put in message
	POP	H		; get start of number string
	PUSH	H
	MVI	B,4		; if 1-4 HI digits are zero, blank them
	MVI	A,'0'
ZBL	CMP	M		; check if first digit is "0"
	JNZ	RETLOC
	MVI	M,0		; if it is, make it null
	INX	H
	DJNZ	ZBL		; if first was "0", check second also
RETLOC: LXI	D,ESCK		; if all "0" leave 1's digit as "0"
	CALL	PMSG		; position cursor
	POP	D
	CALL	PMSG		; print track number
	POP	PSW		; restore the registers
	POP	B
	POP	D
	POP	H
	RET

DECX	PUSH	D		; divide (HL)/(BC) and make one digit quotient
	MVI	E,0		; ASCII and put in memory (DE)
DECLL	ORA	A
	DSBC	B
	INR	E
	JNC	DECLL
	DAD	B		; recover from "overshoot" on last subtraction
	DCR	E
	MOV	A,E		; get quotient
	POP	D
	ORI	'0'		; convert to ASCII
	STAX	D		; put in message
	INX	D		; step to next digit
	RET

INCRE:	LHLD	COUNTR
	INX	H
	SHLD	COUNTR
	LXI	H,CMBFR+1	; INCREMENT TO NEXT TRACK
	MOV	C,M
	INX	H
	MOV	D,M		; BY ADDING S.P.T. TO SECTOR ADDRESS 
	INX	H		; IN COMMAND BUFFER
	MOV	E,M
	LDA	SPT
	MOV	L,A
	MVI	H,0
	DAD	D
	JNC	NOCAR1
	INR	C
NOCAR1: XCHG
	LXI	H,CMBFR+1
	MOV	M,C
	INX	H
	MOV	M,D
	INX	H
	MOV	M,E
	LHLD	MAXTRK 
	LDED	COUNTR
	ANA	A		; TO CLEAR CARRY FOR DSBC
	DSBC	D
	RET			; CARRY IS SET IF NO MORE TRACKS TO CHECK

;
; SUBROUTINES TO TYPE ERROR MESSAGES ON CONSOLE
;
NOHFILE:LXI	D,NOFMSG	; PUT FILE NAME INTO 'NO FILE NAME' MESSAGE
	LXI	H,HEXFILE
	CALL	PUTNAM
	LXI	D,NOFMSG	;  AND TYPE THE MESSAGE
	JMP	EXIT 

USFILE: LXI	D,MENU+11
	LXI	H,HEXFILE
	CALL	PUTNAM
	RET

BDHFIL: LXI	D,BDHMSG	; PUT FILE NAME INTO 'FILE UNRECOGNIZABLE'
	LXI	H,HEXFILE
	CALL	PUTNAM		;  MESSAGE, AND TYPE THE MESSAGE
	LXI	D,BDHMSG


EXIT:	CALL	PMSG		; TYPE OUT ERROR MESSAGE
	JMP	CPM

CLRSCR: LXI	D,ESC$E 	; CLEAR THE SCREEN

PMSG:	MVI	C,MSGOUT
	CALL	BDOS
	RET

;
; SUBROUTINE TO PUT FILE NAME INTO ERROR MESSAGE
;
PUTNAM: MOV	A,M		; GET DRIVE NUMBER
	INX	H
	DCR	A
	JP	NF0
	LDA	4
	ANI	00001111B
NF0:	ADI	'A'		; MAKE IT INTO LETTER
	STAX	D		; PUT LETTER IN MESSAGE
	INX	D		; SKIP OVER ':'
	INX	D
	LXI	B,8		; MOVE FILE NAME TO MESSAGE
	LDIR
	RET

*******************************************************************************

;
; MISCELLANEOUS STORAGE
;
	DS	32		; LOCAL STACK
STACK	DS	0
HEXFILE:DB	0,'        HEX',0,0,0,0
	DB	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
CREC:	DB	0,0,0,0
BADR:	DW	0
COUNT:	DB	0
CHKSUM: DB	0
BYTEC:	DB	0
ADDRS:	DW	0
EOF:	DB	0
CMBFR:	DB	4,0,0,0,0FH,0
STAT:	DB	0
CNUM:	DB	0
BASE$PORT:
	DB	0
COUNTR: DW	0
DOTCNT: DB	0
MAXTRK: DW	0
ERRCNT: DW	0
ILVTBL: DB	00100000B,9,16,48,0  ;LYNX     ;INTERLEAVE FACTOR TABLE
	DB	00000000B,2,46,17,9  ;ST506    ; BYTE 1 = DRIVE/CONTROLLER TYPE
	DB	00000001B,2,136,17,18;CMI5419  ; BYTE 2 = INTERLEAVE FACTOR
	DB	00000010B,2,92,17,13 ;CMI5412  ; BYTE 3 = EXPECTED FORMAT TIME
	DB	00000011B,2,92,17,15 ;MS 10MB  ; BYTE 4 = SECTORS PER TRACK
	DB	01000000B,9,24,0,0   ;Z67    ; BYTE 5 = EXPECTED DISK TEST TIME
	DB	0FFH		   ; MARKS END OF TABLE
SPT:	DB	0		   ; SECTORS PER TRACK

;
;MESSAGES
;
SIGNON: DB	ESC,'E','INITSASI version 1.'
	DW	VERS
	DB	'  (c) Magnolia Microsystems 1982',CR,LF
	DB	'SASI Bus Disk Sub-system Initialization Utility',CR,LF,LF,'$'

NONAME: DB	'No file name specified$'
NOFMSG: DB	'd:filename.HEX not found$'
BDHMSG: DB	'd:filename.HEX is not a recognizable SASI-bus '
	DB	'disk I/O module$'

MENU:	DB	'Using file d:filename.HEX for controller '
CNMMSG: DB	' :                         ',CR,LF,'           '
TXTMSG: DB	'                                                   '
LUNMSG: DB	CR,LF,LF,'Logical Unit '
	DB	'0:                                       ',CR,LF
	DB	'             1:                                       ',CR,LF
	DB	'             2:                                       ',CR,LF
	DB	'             3:                                       ',CR,LF

	DB	LF,'Available Options: 1. Test Controller ',CR,LF
	DB	'                   2. Format and Initialize Drive ',CR,LF
	DB	'Enter Choice:      3. Test Drive  (Drive must be '
	DB	'formatted)',CR,LF
	DB	'                   4. Exit to CP/M',ESC,'Y-.$'

CHCMSG: DB	ESC,'Y-. ',CR,LF
	DB	'               ',CR,LF,ESC,'J',ESC,'Y-.$'
DRVMSG: DB	CR,LF,'TESTING DRIVE ...$'
CONMSG: DB	'TESTING CONTROLLER ...$'
CURPOS: DB	CR,LF,ESC,'j',LF,LF,'$'
QLUMSG: DB	ESC,'kLogical Unit: '
LUN:	DB	'0',BS,ESC,'j$'
NRDYMSG:DB	'  CONTROLLER NOT READY$'
PRTMSG1:DB	'  SW501 SETTING INCORRECT - NO PORT SELECTED$'
PRTMSG2:DB	'  SW501 SETTING DOES NOT MATCH PORT CHOICE$'
NXBMSG: DB	'  Controller does not support software diagnostics$'
BELMSG: DB	ESC,'k ',BS,BEL,'$'
BMSMSG: DB	'  Initialization data for this drive is BAD$'
INSMSG: DB	CR,LF,'INSPECTING DRIVE ...',ESC,'j$'
DNRMSG: DB	'  Drive not ready$'
PREMSG: DB	CR,LF,'  Drive has been previously initialized and may'
	DB	' contain data.',CR,LF
	DB	'  Continuing will DESTROY any data on the drive.',CR,LF
	DB	'  Do you want to DESTROY the data ? (Y or N) N',BS,'$'
NODMSG: DB	'  Drive has not been previously initialized$'
NFMMSG: DB	ESC,'k',ESC,'J',CR,LF,LF,'  Format not attempted',ESC,'J$'
DDSMSG: DB	ESC,'k  Any existing data Destroyed',ESC,'J$'
FMTDSK: DB	CR,LF,'FORMATTING DISK ...$'
TIMMSG: DB	'  Time$'
EXTMSG: DB	' (expect ',ESC,'j$'
DASH:	DB	'-',ESC,'j$'
RPAREN: DB	')$'
COLON:	DB	': ',ESC,'j$'
VERTRK: DB	'  Verifying track ',ESC,'j$'
CURVER: DB	ESC,'Y2^',ESC,'j$'
FBTMSG: DB	CR,LF,'  First bad track: '
TRKNUM: DB	'     $'
LBTMSG: DB	CR,LF,'  Last bad track:  $'
CURLBT: DB	ESC,'Y43',ESC,'j$'
NBTMSG: DB	CR,LF,LF,LF,'  Number of bad tracks: ',ESC,'j$'
WRMSG:	DB	CR,LF,'WRITING INITIALIZATION DATA ...$'
CLRMSG: DB	CR,LF,'CLEARING DIRECTORIES ...$'
WERMSG: DB	'  Write error$'
PMTMSG: DB	CR,LF,LF,'PRESS <RETURN> TO CONTINUE :',ESC,'j$'
ESC$E:	DB	ESC,'E$'
ESCJ:	DB	ESC,'j$'
ESCK:	DB	ESC,'k$'
ESCN:	DB	ESC,'n$'
ESCY:	DB	ESC,'Y$'
CRLF:	DB	CR,LF,'  $'
CRLFLF: DB	CR,LF,LF,'$'
;
; BUFFERS
;
BUFFER: DS	1024		; READ/WRITE BUFFER FOR MAGIC SECTOR
RDBUF:	DS	1024		; READ BUFFER FOR INSPECTING DRIVE
EMTDIR: REPT	1024		; TO CLEAR THE DIRECTORY
	DB	0E5H
	ENDM
	END
